from typing import Iterable, Optional, Callable

from redis import Redis


def read_list(
    redis_conn: "Redis[bytes]",
    key: str,
    *,
    chunk_size: int = 4096,
    filter_fn: Optional[Callable[[bytes], bool]] = None,
    last_n: Optional[int] = None,
) -> Iterable[bytes]:
    """
    Read a possibly large Redis list in chunks.

    Avoids OOMs on the Redis side.

    :param redis_conn: Redis connection
    :param key: Key
    :param chunk_size: How many lines to read per request.
    :param filter_fn: Only include lines that pass this filter
    :param last_n: How many last lines to return, filtering is applied before limiting.
    :return:
    """
    # LLEN returns zero for a non-existent key,
    # and there would be nothing to iterate from an empty
    # list anyway, so we can do one query less.
    list_len = redis_conn.llen(key)
    if list_len == 0:
        return

    if chunk_size <= 0:
        chunk_size = 4096

    results = []

    # Regarding that - 1 there, see this from https://redis.io/commands/lrange:
    # > Note that if you have a list of numbers from 0 to 100, LRANGE list 0 10
    # > will return 11 elements, that is, the rightmost item is included.
    end_index = list_len - 1
    remaining_needed = last_n if last_n and last_n > 0 else list_len

    # read until we have enough items, or we run out of the list
    while end_index >= 0 and remaining_needed > 0:
        start_index = max(0, end_index - chunk_size + 1)

        chunk = redis_conn.lrange(key, start_index, end_index)
        if not chunk:
            break

        for item in reversed(chunk):
            if not filter_fn or filter_fn(item):
                results.append(item)
                remaining_needed -= 1
                if remaining_needed == 0:
                    break

        # move the reading window further back
        end_index = start_index - 1

    yield from reversed(results)



import time
from typing import Any, Callable, Dict, Optional, Union

from redis import Redis

import minique.encoding as encoding
from minique.enums import JobStatus
from minique.excs import DuplicateJob
from minique.models.job import Job
from minique.models.queue import Queue
from minique.models.priority_queue import PriorityQueue
from minique.utils import get_random_pronounceable_string


def enqueue(
    redis: "Redis[bytes]",
    queue_name: str,
    callable: Union[Callable[..., Any], str],
    kwargs: Optional[Dict[str, Any]] = None,
    job_id: Optional[str] = None,
    job_ttl: int = 0,
    result_ttl: int = 86400 * 7,
    encoding_name: Optional[str] = None,
) -> Job:
    """
    Queue up callable as a job.

    :param redis: Redis connection
    :param queue_name: Name of the queue to enqueue the job in.
    :param callable: A dotted path to the callable to execute on the worker.
    :param kwargs: Keyword arguments to pass to the callable.
    :param job_id: An identifier for the job; defaults to a random string.
    :param job_ttl: Time-to-live for the job in seconds; defaults to never expire.
    :param result_ttl: Time-to-live for the result in seconds; defaults to 7 days.
    :param encoding_name: Name of the encoding to use for the job payload; defaults to JSON.
    :raises minique.excs.DuplicateJob: If a job with the same ID already exists.
    :raises minique.excs.NoSuchJob: If the job does not exist right after creation.
    """
    queue = Queue(redis, queue_name)
    job = _define_and_store_job(
        redis=redis,
        callable=callable,
        kwargs=kwargs,
        job_id=job_id,
        job_ttl=job_ttl,
        result_ttl=result_ttl,
        encoding_name=encoding_name,
        queue=queue,
    )
    return job


def enqueue_priority(
    redis: "Redis[bytes]",
    queue_name: str,
    callable: Union[Callable[..., Any], str],
    kwargs: Optional[Dict[str, Any]] = None,
    job_id: Optional[str] = None,
    job_ttl: int = 0,
    result_ttl: int = 86400 * 7,
    encoding_name: Optional[str] = None,
    priority: int = 0,
) -> Job:
    """
    Queue up callable as a job, placing the job at the last place for its priority.

    :param redis: Redis connection
    :param queue_name: Name of the queue to enqueue the job in.
    :param callable: A dotted path to the callable to execute on the worker.
    :param kwargs: Keyword arguments to pass to the callable.
    :param job_id: An identifier for the job; defaults to a random string.
    :param job_ttl: Time-to-live for the job in seconds; defaults to never expire.
    :param result_ttl: Time-to-live for the result in seconds; defaults to 7 days.
    :param encoding_name: Name of the encoding to use for the job payload; defaults to JSON.
    :param priority: Priority number of this job, defaults to zero.
    :raises minique.excs.DuplicateJob: If a job with the same ID already exists.
    :raises minique.excs.NoSuchJob: If the job does not exist right after creation.
    """
    queue = PriorityQueue(redis, queue_name)
    job = _define_and_store_job(
        redis=redis,
        callable=callable,
        kwargs=kwargs,
        job_id=job_id,
        job_ttl=job_ttl,
        result_ttl=result_ttl,
        encoding_name=encoding_name,
        queue=queue,
        priority=priority,
    )
    return job


def store(
    redis: "Redis[bytes]",
    callable: Union[Callable[..., Any], str],
    kwargs: Optional[Dict[str, Any]] = None,
    job_id: Optional[str] = None,
    job_ttl: int = 0,
    result_ttl: int = 86400 * 7,
    encoding_name: Optional[str] = None,
) -> Job:
    """
    Store callable as a job without placing it in the queue.

    :param redis: Redis connection
    :param callable: A dotted path to the callable to execute on the worker.
    :param kwargs: Keyword arguments to pass to the callable.
    :param job_id: An identifier for the job; defaults to a random string.
    :param job_ttl: Time-to-live for the job in seconds; defaults to never expire.
    :param result_ttl: Time-to-live for the result in seconds; defaults to 7 days.
    :param encoding_name: Name of the encoding to use for the job payload; defaults to JSON.
    :raises minique.excs.DuplicateJob: If a job with the same ID already exists.
    :raises minique.excs.NoSuchJob: If the job does not exist right after creation.
    """
    return _define_and_store_job(
        redis=redis,
        callable=callable,
        kwargs=kwargs,
        job_id=job_id,
        job_ttl=job_ttl,
        result_ttl=result_ttl,
        encoding_name=encoding_name,
    )


def get_job(
    redis: "Redis[bytes]",
    job_id: str,
) -> Job:
    job = Job(redis, job_id)
    job.ensure_exists()
    return job


def cancel_job(
    redis: "Redis[bytes]",
    job_id: str,
    expire_time: Optional[int] = None,
) -> bool:
    """
    Cancel the job with the given job ID.

    If a worker is already busy with the job, it may not immediately quit,
    and as such, the job is not set to canceled state.

    You can optionally specify an expiration time for the job object; this overrides
    any `job_ttl` set in the job payload itself.  The assumption is that you don't
    need a canceled job object to hang around forever.

    :param redis: Redis connection
    :param job_id: Job ID.
    :param expire_time: Expiration time for the job object in seconds.
    :raises minique.excs.NoSuchJob: If the job does not exist.
    """
    job = get_job(redis, job_id)
    if not (job.has_finished or job.has_started):
        with redis.pipeline() as p:
            p.hset(job.redis_key, "status", JobStatus.CANCELLED.value)
            queue_name = job.get_queue_name()
            if queue_name:
                job.dequeue()
            if expire_time:
                p.expire(job.redis_key, expire_time)
            p.execute()
        return True
    return False


def _define_and_store_job(
    *,
    redis: "Redis[bytes]",
    callable: Union[Callable[..., Any], str],
    kwargs: Optional[Dict[str, Any]] = None,
    job_id: Optional[str] = None,
    job_ttl: int = 0,
    result_ttl: int = 86400 * 7,
    encoding_name: Optional[str] = None,
    queue: Optional[Queue] = None,
    priority: Optional[int] = None,
) -> Job:
    if not encoding_name:
        encoding_name = encoding.default_encoding_name
    encoder = encoding.registry[str(encoding_name)]()

    if not isinstance(callable, str):
        callable = f"{callable.__module__}.{callable.__qualname__}"

    if not job_id:
        job_id = get_random_pronounceable_string()

    job = Job(redis, job_id)
    if job.exists:
        raise DuplicateJob(f"duplicate job: {job_id}")

    payload: Dict[str, Any] = {
        "callable": str(callable),
        "kwargs": encoder.encode(kwargs or {}),
        "encoding_name": encoding_name,
        "status": JobStatus.NONE.value,
        "ctime": time.time(),
        "job_ttl": int(job_ttl),
        "result_ttl": int(result_ttl),
    }
    if queue:
        payload["queue"] = queue.name
    if priority is not None:
        payload["priority"] = priority

    with redis.pipeline() as p:
        p.hset(job.redis_key, mapping=payload)  # type: ignore[arg-type]
        if payload["job_ttl"] > 0:
            p.expire(job.redis_key, payload["job_ttl"])
        p.execute()

    if queue:
        queue.add_job(job)

    job.ensure_exists()
    return job



from typing import TYPE_CHECKING

from redis import Redis

from minique.models.queue import Queue

if TYPE_CHECKING:
    from minique.models.job import Job


ADD_JOB_SCRIPT = """
local queue_key = ARGV[1]
local queue_prio_lookup = ARGV[2]
local job_key = ARGV[3]
local job_id = ARGV[4]
local job_priority = tonumber(redis.call("HGET", job_key, "priority")) or 0
local job_index = 1
local insert_status = -1

-- Check the priority for the last job in the queue, we might become the new last job
-- This is generally more efficient than iterating the queue.
local last_job_id = redis.call("LRANGE", queue_key, -1, -1)[1]
if last_job_id ~= nil then
    local last_job_priority = tonumber(redis.call("HGET", queue_prio_lookup, last_job_id)) or 0
    if last_job_priority >= job_priority then
        insert_status = redis.call("RPUSH", queue_key, job_id)
        -- Insert status is the length of the queue
        job_index = insert_status
    end
else
    -- The queue is empty.
    insert_status = redis.call("RPUSH", queue_key, job_id)
end

if insert_status == -1 then
    -- Our place in the queue is somewhere in the middle, search for it
    for i, queued_job_id in ipairs(redis.call("LRANGE", queue_key, 0, -1)) do
        local queued_job_priority = tonumber(redis.call("HGET", queue_prio_lookup, queued_job_id)) or 0
        if queued_job_priority < job_priority then
            -- Found the first job with a lower priority, insert before it
            insert_status = redis.call("LINSERT", queue_key, "BEFORE", queued_job_id, job_id)
            job_index = i
            break
        end
    end
end

if insert_status == -1 then
    -- No higher priority jobs found, insert to the end
    redis.call("RPUSH", queue_key, job_id)
end

-- Introduce the newly added job to the priority lookup hash
redis.call("HSET", queue_prio_lookup, job_id, job_priority)

return job_index
"""

PRIO_HASH_CLEANER_SCRIPT = """
local queue_key = ARGV[1]
local queue_prio_lookup = ARGV[2]

local queue_len = tonumber(redis.call("LLEN", queue_key))
local hash_len = tonumber(redis.call("HLEN", queue_prio_lookup))

-- Tolerate some excess
if queue_len + 100 >= hash_len then
    return -1
end

local new_hash = {}
local count = 0

for i, queued_job_id in ipairs(redis.call("LRANGE", queue_key, 0, -1)) do
    local priority = redis.call("HGET", queue_prio_lookup, queued_job_id)
    new_hash[queued_job_id] = priority
    count = i
end
redis.call("DEL", queue_prio_lookup)

for queued_job_id, priority in pairs(new_hash) do
    redis.call("HSET", queue_prio_lookup, queued_job_id, priority)
end

return hash_len - count
"""


class PriorityQueue(Queue):
    """Alternate Queue implementation that uses the optional `Job.priority` attribute
    to sort the queue when adding jobs.

    When using the priority queue, ensure the `PriorityQueue.finish_job()` method is called
    for each job after it leaves the queue to trim the job priority lookup hash.
    This can be performed by either the consumer or the job manager, or both.

    Alternately, or in addition, periodically call `PriorityQueue.periodic_clean()` to
    remove stale keys from the priority lookup hash.
    """

    def __init__(self, redis: "Redis[bytes]", name: str):
        super().__init__(redis, name)
        self.add_job_script = redis.register_script(ADD_JOB_SCRIPT)
        self.hash_clean_script = redis.register_script(PRIO_HASH_CLEANER_SCRIPT)

    @property
    def prio_key(self) -> str:
        return f"{self.redis_key}prio"

    def add_job(self, job: "Job") -> int:
        script_response = self.add_job_script(
            keys=[self.redis_key, self.prio_key, job.redis_key],
            args=[self.redis_key, self.prio_key, job.redis_key, job.id],
        )
        return int(script_response) - 1

    def dequeue_job(self, job_id: str) -> bool:
        """Dequeue a job from any position in the queue, cleaning it from the priority
        lookup hash.
        """
        self.redis.hdel(self.prio_key, job_id)
        num_removed = self.redis.lrem(self.redis_key, 0, job_id)
        return num_removed > 0

    def clean_job(self, job: "Job") -> None:
        """Cleans up job data after the job has exited the queue."""
        self.redis.hdel(self.prio_key, job.id)

    def periodic_clean(self) -> int:
        """Perform occasional maintenance on the data structures

        :return: Number of cleaned up values
        """
        script_response = self.hash_clean_script(
            keys=[self.redis_key, self.prio_key],
            args=[self.redis_key, self.prio_key],
        )
        return int(script_response)
